import getpass
import os
import platform
import random
import socket
import string
import sys
from datetime import datetime
from typing import Any, Dict, Union

from .api import Novem401, NovemAPI
from .setup import setup
from .utils import check_if_profile_exists, update_config


class cl:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


def disable_colors() -> None:
    cl.HEADER = ""
    cl.OKBLUE = ""
    cl.OKCYAN = ""
    cl.OKGREEN = ""
    cl.WARNING = ""
    cl.FAIL = ""
    cl.ENDC = ""
    cl.BOLD = ""
    cl.UNDERLINE = ""


def colors() -> None:
    # disable colors if not supported
    for handle in [sys.stdout, sys.stderr]:
        if (hasattr(handle, "isatty") and handle.isatty()) or (
            "TERM" in os.environ and os.environ["TERM"] == "ANSI"
        ):
            if platform.system() == "Windows" and not (
                "TERM" in os.environ and os.environ["TERM"] == "ANSI"
            ):
                disable_colors()
        else:
            disable_colors()


def do_update_config(
    profile: str, username: str, api_root: str, token_name: str, token: str
) -> None:
    (status, path) = update_config(
        profile, username, api_root, token_name, token
    )

    print(
        f"{cl.OKGREEN} \u2713 {cl.ENDC}new token "
        f'{cl.OKCYAN}"{token_name}"{cl.ENDC} created and'
        f" saved to {path}"
    )

    # save file


def init_config(args: Dict[str, str] = None) -> None:
    """
    Initialize the novem config and write to file.
    """

    if not args:
        # make mypy happy, return if no argument supplied
        return

    token_name: Union[str, None] = None

    if "token" in args:
        token_name = args["token"]

    api_root: str = args["api-url"]
    profile: str = args["profile-name"]
    force: str = args["force"]

    # first check if we have a valid config
    profile_exists: bool = check_if_profile_exists(profile)
    if profile_exists and not force:
        print(
            f"{cl.WARNING} ! {cl.ENDC}"
            f' The supplied profile "{cl.OKCYAN}{profile}{cl.ENDC}" already '
            f"exist, use --force to override"
        )
        sys.exit(1)

    valid_char_sm = string.ascii_lowercase + string.digits
    valid_char = valid_char_sm + "-_"
    hostname: str = socket.gethostname()

    if not token_name:
        token_hostname: str = "".join(
            [x for x in hostname.lower() if x in valid_char]
        )
        nounce: str = "".join(random.choice(valid_char_sm) for _ in range(8))
        token_name = f"novem-python-{token_hostname}-{nounce}"

    new_token_name = "".join([x for x in token_name if x in valid_char])

    if token_name != new_token_name:
        print(
            f"{cl.WARNING} ! {cl.ENDC}"
            "The supplied token contained invalid charracters,"
            f' token changed to "{cl.OKCYAN}{new_token_name}{cl.ENDC}"'
        )
        token_name = new_token_name

    # get novem username
    username = input(" \u2022 novem.no username: ")
    # username = "abc"

    # get novem password
    password = getpass.getpass(" \u2022 novem.no password: ")

    # authenticate and request token by name
    req = {
        "username": username,
        "password": password,
        "token_name": token_name,
        "token_description": (
            f'cli token created for "{hostname}" '
            f'on "{datetime.now():%Y-%m-%d:%H:%M:%S}"'
        ),
    }

    # let's grab our token
    novem = NovemAPI(
        api_root=api_root,
        ignore_config=True,
    )

    try:
        token = novem.create_token(req)
    except Novem401:
        print("Invalid username and/or password")
        sys.exit(1)

    # let's write our config
    do_update_config(profile, username, api_root, token_name, token)


def run_cli(raw_args: Any = None) -> None:
    colors()

    # Gather the provided arguements as an array.
    args: Dict[str, str] = setup(raw_args)

    if args and args["init"]:
        # init_config(args["token"], args["api"])
        init_config(args)


__all__ = ["run_cli"]



from typing import Any, Dict, Optional

import requests

from ..api import Novem403, Novem404, NovemAPI
from ..version import __version__
from .shared import NovemShare


class NovemVisAPI(NovemAPI):
    """ """

    hdr_post: Dict[str, str] = {
        "User-Agent": f"NovemPythonLibrary-{__version__}",
        "Content-type": "text/plain",
    }
    hdr_put: Dict[str, str] = {
        "User-Agent": f"NovemPythonLibrary-{__version__}",
    }
    hdr_get: Dict[str, str] = {
        "User-Agent": f"NovemPythonLibrary-{__version__}",
    }
    hdr_del: Dict[str, str] = {
        "User-Agent": f"NovemPythonLibrary-{__version__}",
    }

    shared: Optional[NovemShare] = None

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

        self.shared = NovemShare(self)

    def __setattr__(self, name: str, value: Any) -> None:
        if name == "shared" and self.shared:
            self.shared.set(value)
        else:
            super().__setattr__(name, value)

    def api_read(self, relpath: str) -> str:
        """
        Read the api value located at realtive path
        """

        r = requests.get(
            f"{self.api_root}vis/plots/{self.id}{relpath}",
            auth=("", self.token),
            headers=self.hdr_get,
        )

        # TODO: verify result and raise exception if not ok
        if r.status_code == 404:
            raise Novem404

        if r.status_code == 403:
            raise Novem403

        return r.text

    def api_delete(self, relpath: str) -> None:
        """
        relpath: relative path to the plot baseline /config/type
                 for the type file in the config folder
        value: the value to write to the file
        """

        r = requests.delete(
            f"{self.api_root}vis/plots/{self.id}{relpath}",
            auth=("", self.token),
            headers=self.hdr_del,
        )

        if r.status_code == 404:
            raise Novem404

        if r.status_code == 403:
            raise Novem403

        # TODO: verify result and raise exception if not ok
        if not r.ok:
            print(r)
            print(r.text)
            print("should raise an error")

    def api_create(self, relpath: str) -> None:
        """
        relpath: relative path to the plot baseline /config/type
                 for the type file in the config folder
        value: the value to write to the file
        """

        path = f"{self.api_root}vis/plots/{self.id}{relpath}"
        r = requests.put(
            path,
            auth=("", self.token),
            headers=self.hdr_put,
        )

        if r.status_code == 404:
            raise Novem404(path)

        if r.status_code == 403:
            raise Novem403(path)

        if r.status_code == 409:
            # we will ignore 409 errors
            # as creating objects that already exist is not a problem
            return

        # TODO: verify result and raise exception if not ok

        if not r.ok:
            print(r)
            print(r.text)
            print("should raise a general error")

    def api_write(self, relpath: str, value: str) -> None:
        """
        relpath: relative path to the plot baseline /config/type
                 for the type file in the config folder
        value: the value to write to the file
        """

        r = requests.post(
            f"{self.api_root}vis/plots/{self.id}{relpath}",
            auth=("", self.token),
            headers=self.hdr_post,
            data=value,
        )
        # TODO: verify result and raise exception if not ok
        if not r.ok:
            print(r)
            print(r.text)
            print("should raise an error")

        if r.status_code == 404:
            raise Novem404

        if r.status_code == 403:
            raise Novem403

        # TODO: verify result and raise exception if not ok
        if not r.ok:
            print(r)
            print(r.text)
            print("should raise an error")



import configparser
import sys
from typing import Any, Dict, Optional

import requests

from .utils import get_config_path


class NovemException(Exception):
    pass


class Novem404(NovemException):
    pass


class Novem403(NovemException):
    pass


class Novem401(NovemException):
    pass


class NovemAPI(object):
    """
    Novem API class

    * Read config file
    * Communicate with api.novem.no
    * Offer utilities for subclasses
    """

    id: Optional[str] = None

    def __init__(self, **kwargs: Any) -> None:
        """"""

        if "config_path" not in kwargs:
            (novem_dir, config_path) = get_config_path()
        else:
            config_path = kwargs["config_path"]

        ignore_config = False
        if "ignore_config" in kwargs:
            ignore_config = kwargs["ignore_config"]

        config = configparser.ConfigParser()

        # check if novem config file exist
        config.read(config_path)

        try:
            self.token = config["default"]["token"]
            self.api_root = config["default"]["api_root"]
        except KeyError:
            if not ignore_config:
                print("Novem config file is missing.")
                print(
                    "Either specify config file location with "
                    "the config_path parameter."
                )
                print("or setup a new token using python -m novem --init")
                sys.exit(0)
            else:
                if "default" not in config:
                    self.api_root = kwargs["api_root"]
                pass

        if self.api_root[-1] != "/":
            # our code assumes that the api_root ends with a /
            self.api_root = f"{self.api_root}/"

    def parse_kwargs(self, **kwargs: Any) -> None:
        """
        Parse the arguments and invoke the novem api
        """

        if "type" in kwargs:
            self.type = kwargs["type"]
        if "api_root" in kwargs:
            self.api_root = kwargs["api_root"]

    def create_token(self, params: Dict[str, str]) -> str:

        r = requests.post(
            f"{self.api_root}/token",
            json=params,
        )

        if not r.ok:
            resp = r.json()
            if r.status_code == 401:
                raise Novem401(resp["message"])

        res = r.json()

        return res["token"]

    def api_read(self, relpath: str) -> str:
        """
        Read the api value located at realtive path
        """
        pass

    def api_delete(self, relpath: str) -> None:
        """
        relpath: relative path to the plot baseline /config/type
                 for the type file in the config folder
        value: the value to write to the file
        """
        pass

    def api_create(self, relpath: str) -> None:
        """
        relpath: relative path to the plot baseline /config/type
                 for the type file in the config folder
        value: the value to write to the file
        """
        pass

    def api_write(self, relpath: str, value: str) -> None:
        """
        relpath: relative path to the plot baseline /config/type
                 for the type file in the config folder
        value: the value to write to the file
        """
        pass



from typing import Any

from novem.vis import NovemVisAPI


class Plot(NovemVisAPI):
    """
    Novem plot class
    """

    def __init__(self, id: str, **kwargs: Any) -> None:
        """
        :name plot name, duplicate entry will update the plot

        :type the type of plot
        :caption caption of the plot
        :title title of the plot
        """
        self.id = id

        super().__init__(**kwargs)

        # let's create our plot
        self.api_create("")

        # if information neccessary to
        # terminate and print
        self.parse_kwargs(**kwargs)

    def __call__(self, data: Any, **kwargs: Any) -> Any:
        """
        Set's the data of the plot

        The paramter either needs to be a text string
        of CSV formatted text or an object with a to_csv
        function.
        """

        raw_str: str = ""
        to_csv = getattr(data, "to_csv", None)

        if callable(to_csv):
            # if data has a callable csv funciton we'll
            # assume it's a pandas dataframe
            raw_str = data.to_csv()
        else:
            raw_str = str(data)

        # setting the data object will invoke a server write
        self.data = raw_str

        # also update our chart varibales
        self.parse_kwargs(**kwargs)

        # return the original object so users can chain the dataframe
        return data

    # we'll implement generic properties common across all plots here
    @property
    def type(self) -> str:
        return self.api_read("/config/type").strip()

    @type.setter
    def type(self, value: str) -> None:
        return self.api_write("/config/type", value)

    @property
    def name(self) -> str:
        return self.api_read("/name").strip()

    @name.setter
    def name(self, value: str) -> None:
        return self.api_write("/name", value)

    @property
    def description(self) -> str:
        return self.api_read("/description")

    @description.setter
    def description(self, value: str) -> None:
        return self.api_write("/description", value)

    @property
    def data(self) -> str:
        return self.api_read("/data")

    @data.setter
    def data(self, value: str) -> None:
        return self.api_write("/data", value)

    @property
    def url(self) -> str:
        return self.api_read("/url").strip()



import configparser
import os
import stat
from os import path
from pathlib import Path
from typing import Tuple, Union

NOVEM_PATH = "novem"
NOVEM_NAME = "novem.conf"


def get_user_config_directory() -> Union[str, None]:
    """Returns a platform-specific root directory for user config settings."""
    # On Windows, prefer %LOCALAPPDATA%, then %APPDATA%, since we can expect
    # the AppData directories to be ACLed to be visible only to the user and
    # admin users (https://stackoverflow.com/a/7617601/1179226). If neither is
    # set, return None instead of falling back to something that may be
    # world-readable.
    if os.name == "nt":
        appdata = os.getenv("LOCALAPPDATA")
        if appdata:
            return appdata
        appdata = os.getenv("APPDATA")
        if appdata:
            return appdata
        return None
    # On non-windows, use XDG_CONFIG_HOME if set, else default to ~/.config.
    xdg_config_home = os.getenv("XDG_CONFIG_HOME")
    if xdg_config_home:
        return xdg_config_home
    return os.path.join(os.path.expanduser("~"), ".config")


def get_config_path() -> Tuple[str, str]:
    """
    Get default configuration path
    """

    config_path: Union[str, None] = get_user_config_directory()
    novem_dir = f"{config_path}/{NOVEM_PATH}"
    novem_config = f"{config_path}/{NOVEM_PATH}/{NOVEM_NAME}"

    return (novem_dir, novem_config)


def update_config(
    profile: str, username: str, api_root: str, token_name: str, token: str
) -> Tuple[bool, str]:
    """
    Write configuration to file
    """

    (novem_dir, novem_config) = get_config_path()

    # create path and file if not exist
    Path(novem_dir).mkdir(parents=True, exist_ok=True)
    Path(novem_config).touch(mode=stat.S_IRUSR | stat.S_IWUSR, exist_ok=True)
    os.chmod(
        novem_config, stat.S_IRUSR | stat.S_IWUSR
    )  # ensure file is not world readable

    config = configparser.ConfigParser()

    # read our config object
    config.read(novem_config)

    # add/update our section
    if not config.has_section(profile):
        config.add_section(profile)

    config.set(profile, "username", username)
    config.set(profile, "api_root", api_root)
    config.set(profile, "token_name", token_name)
    config.set(profile, "token", token)

    with open(novem_config, "w+") as configfile:
        config.write(configfile)

    return (True, novem_config)


def check_if_profile_exists(profile: str) -> bool:
    """
    Check if config dir already contains a valid token for the profile
    """

    (novem_dir, novem_config) = get_config_path()

    # check if path exists
    if not path.exists(novem_config):
        # profile don't exist

        return False

    config = configparser.ConfigParser()

    # read our config object
    config.read(novem_config)

    # add/update our section
    if config.has_section(profile):
        return True

    return False



import argparse as ap
import shutil
from typing import Any, Dict

width = min(80, shutil.get_terminal_size().columns - 2)


def formatter(prog: str) -> ap.RawDescriptionHelpFormatter:
    return ap.RawDescriptionHelpFormatter(prog, width=width)


def setup(raw_args: Any = None) -> Dict[str, str]:
    # formatter = lambda prog: ap.RawDescriptionHelpFormatter(prog,
    # width=width)

    parser = ap.ArgumentParser(
        prog="novem",
        description="Novem commandline interface.",
        formatter_class=formatter,
    )

    general = parser.add_argument_group("general")

    general.add_argument(
        "--api-url",
        dest="api-url",
        action="store",
        required=False,
        default="https://api.novem.no/v1",
        help="api url to use, defaul is https://api.novem.no/v1",
    )

    general.add_argument(
        "--profile-name",
        dest="profile-name",
        action="store",
        required=False,
        default="default",
        help=(
            "Which profile to use, combine with --init to setup a"
            " new profile and --force to override an existing token"
        ),
    )

    general.add_argument(
        "--token",
        dest="token",
        action="store",
        required=False,
        default=None,
        help="Use this token instead, overrides profile lookup",
    )

    setup = parser.add_argument_group("setup")

    setup.add_argument(
        "--init",
        dest="init",
        action="store_true",
        help="authenticate with the novem service and create"
        " default configuration",
    )

    setup.add_argument(
        "--force",
        dest="force",
        action="store_true",
        required=False,
        help="force reinit of existing profile",
    )

    setup.add_argument(
        "--token_name",
        dest="token_name",
        action="store",
        required=False,
        default=None,
        help="name of token (lowercase alphanumeric, no whitespace)",
    )

    # Gather the provided arguements as an array.
    args: Dict[str, str] = vars(parser.parse_args(raw_args))

    return args
